@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2018 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2020 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -8,17 +8,19 @@
 #define MEMFILE_LEASE_MGR_H
 
 #include <asiolink/interval_timer.h>
+#include <database/database_connection.h>
 #include <dhcp/hwaddr.h>
 #include <dhcpsrv/csv_lease_file4.h>
 #include <dhcpsrv/csv_lease_file6.h>
 #include <dhcpsrv/memfile_lease_storage.h>
-#include <dhcpsrv/database_connection.h>
 #include <dhcpsrv/lease_mgr.h>
 #include <util/process_spawn.h>
 
 #include <boost/scoped_ptr.hpp>
 #include <boost/shared_ptr.hpp>
 
+#include <mutex>
+
 namespace isc {
 namespace dhcp {
 
@@ -37,7 +39,8 @@ class LFCSetup;
 /// in a significant growth of the lease file size over time, because the lease
 /// file will contain many entries for each lease. In order to mitigate this
 /// problem, the backend implements the Lease File Cleanup mechanism which is
-/// described on the Kea wiki: http://kea.isc.org/wiki/LFCDesign.
+/// described on the Kea wiki:
+/// https://gitlab.isc.org/isc-projects/kea/wikis/designs/Lease-File-Cleanup-design.
 ///
 /// The backend installs an @c asiolink::IntervalTimer to periodically execute
 /// the @c Memfile_LeaseMgr::lfcCallback. This callback function controls
@@ -71,8 +74,8 @@ class LFCSetup;
 /// parameter in the database access string. The [path] is the
 /// absolute path to the file (including file name). If this parameter
 /// is not specified, the default location in the installation
-/// directory is used: var/kea/kea-leases4.csv and
-/// var/kea/kea-leases6.csv.
+/// directory is used: <install-dir>/var/lib/kea/kea-leases4.csv and
+/// <install-dir>/var/lib/kea/kea-leases6.csv.
 class Memfile_LeaseMgr : public LeaseMgr {
 public:
 
@@ -83,12 +86,13 @@ public:
     /// Version history:
     /// 1.0 - initial version (released in Kea 0.9)
     /// 2.0 - hwaddr column added (to be released in Kea 0.9.1)
+    /// 2.1 - user context column add (to be released in Kea 1.5)
     ///
     /// @{
     static const int MAJOR_VERSION = 2;
 
     /// Defines minor version of the memfile backend.
-    static const int MINOR_VERSION = 0;
+    static const int MINOR_VERSION = 1;
 
     /// @}
 
@@ -129,7 +133,7 @@ public:
     ///
     /// @param parameters A data structure relating keywords and values
     ///        concerned with the database.
-    Memfile_LeaseMgr(const DatabaseConnection::ParameterMap& parameters);
+    Memfile_LeaseMgr(const db::DatabaseConnection::ParameterMap& parameters);
 
     /// @brief Destructor (closes file)
     virtual ~Memfile_LeaseMgr();
@@ -140,11 +144,15 @@ public:
     /// @brief Adds an IPv4 lease.
     ///
     /// @param lease lease to be added
+    ///
+    /// @result true if the lease was added, false if not
     virtual bool addLease(const Lease4Ptr& lease);
 
     /// @brief Adds an IPv6 lease.
     ///
     /// @param lease lease to be added
+    ///
+    /// @result true if the lease was added, false if not
     virtual bool addLease(const Lease6Ptr& lease);
 
     /// @brief Returns existing IPv4 lease for specified IPv4 address.
@@ -154,7 +162,7 @@ public:
     ///
     /// @param addr An address of the searched lease.
     ///
-    /// @return a collection of leases
+    /// @return a pointer to the lease (or NULL if a lease is not found)
     virtual Lease4Ptr getLease4(const isc::asiolink::IOAddress& addr) const;
 
     /// @brief Returns existing IPv4 leases for specified hardware address.
@@ -188,6 +196,8 @@ public:
     /// @brief Returns existing IPv4 lease for specified client-id
     ///
     /// @param client_id client identifier
+    ///
+    /// @return lease collection
     virtual Lease4Collection getLease4(const ClientId& client_id) const;
 
     /// @brief Returns IPv4 lease for specified client-id/hwaddr/subnet-id tuple
@@ -227,11 +237,46 @@ public:
     /// @return Lease collection (may be empty if no IPv4 lease found).
     virtual Lease4Collection getLeases4(SubnetID subnet_id) const;
 
+    /// @brief Returns all IPv4 leases for the particular hostname.
+    ///
+    /// @param hostname hostname in lower case.
+    ///
+    /// @return Lease collection (may be empty if no IPv4 lease found).
+    virtual Lease4Collection getLeases4(const std::string& hostname) const;
+
     /// @brief Returns all IPv4 leases.
     ///
     /// @return Lease collection (may be empty if no IPv4 lease found).
     virtual Lease4Collection getLeases4() const;
 
+    /// @brief Returns range of IPv4 leases using paging.
+    ///
+    /// This method implements paged browsing of the lease database. The first
+    /// parameter specifies a page size. The second parameter is optional and
+    /// specifies the starting address of the range. This address is excluded
+    /// from the returned range. The IPv4 zero address (default) denotes that
+    /// the first page should be returned. There is no guarantee about the
+    /// order of returned leases.
+    ///
+    /// The typical usage of this method is as follows:
+    /// - Get the first page of leases by specifying IPv4 zero address as the
+    ///   beginning of the range.
+    /// - Last address of the returned range should be used as a starting
+    ///   address for the next page in the subsequent call.
+    /// - If the number of leases returned is lower than the page size, it
+    ///   indicates that the last page has been retrieved.
+    /// - If there are no leases returned it indicates that the previous page
+    ///   was the last page.
+    ///
+    /// @param lower_bound_address IPv4 address used as lower bound for the
+    /// returned range.
+    /// @param page_size maximum size of the page returned.
+    ///
+    /// @return Lease collection (may be empty if no IPv4 lease found).
+    virtual Lease4Collection
+    getLeases4(const asiolink::IOAddress& lower_bound_address,
+               const LeasePageSize& page_size) const;
+
     /// @brief Returns existing IPv6 lease for a given IPv6 address.
     ///
     /// This function returns a copy of the lease. The modification in the
@@ -240,7 +285,7 @@ public:
     /// @param type specifies lease type: (NA, TA or PD)
     /// @param addr An address of the searched lease.
     ///
-    /// @return smart pointer to the lease (or NULL if a lease is not found)
+    /// @return a pointer to the lease (or NULL if a lease is not found)
     virtual Lease6Ptr getLease6(Lease::Type type,
                                 const isc::asiolink::IOAddress& addr) const;
 
@@ -253,7 +298,8 @@ public:
     ///
     /// @return collection of IPv6 leases
     virtual Lease6Collection getLeases6(Lease::Type type,
-                                        const DUID& duid, uint32_t iaid) const;
+                                        const DUID& duid,
+                                        uint32_t iaid) const;
 
     /// @brief Returns existing IPv6 lease for a given DUID + IA + subnet-id +
     /// lease type combination.
@@ -267,7 +313,8 @@ public:
     /// @param subnet_id identifier of the subnet the lease must belong to
     ///
     /// @return lease collection (may be empty if no lease is found)
-    virtual Lease6Collection getLeases6(Lease::Type type, const DUID& duid,
+    virtual Lease6Collection getLeases6(Lease::Type type,
+                                        const DUID& duid,
                                         uint32_t iaid,
                                         SubnetID subnet_id) const;
 
@@ -278,11 +325,51 @@ public:
     /// @return Lease collection (may be empty if no IPv6 lease found).
     virtual Lease6Collection getLeases6(SubnetID subnet_id) const;
 
+    /// @brief Returns all IPv6 leases for the particular hostname.
+    ///
+    /// @param hostname hostname in lower case.
+    ///
+    /// @return Lease collection (may be empty if no IPv6 lease found).
+    virtual Lease6Collection getLeases6(const std::string& hostname) const;
+
     /// @brief Returns all IPv6 leases.
     ///
     /// @return Lease collection (may be empty if no IPv6 lease found).
     virtual Lease6Collection getLeases6() const;
 
+    /// @brief Returns IPv6 leases for the DUID.
+    ///
+    /// @param duid client DUID
+    virtual Lease6Collection getLeases6(const DUID& duid) const;
+
+    /// @brief Returns range of IPv6 leases using paging.
+    ///
+    /// This method implements paged browsing of the lease database. The first
+    /// parameter specifies a page size. The second parameter is optional and
+    /// specifies the starting address of the range. This address is excluded
+    /// from the returned range. The IPv6 zero address (default) denotes that
+    /// the first page should be returned. There is no guarantee about the
+    /// order of returned leases.
+    ///
+    /// The typical usage of this method is as follows:
+    /// - Get the first page of leases by specifying IPv6 zero address as the
+    ///   beginning of the range.
+    /// - Last address of the returned range should be used as a starting
+    ///   address for the next page in the subsequent call.
+    /// - If the number of leases returned is lower than the page size, it
+    ///   indicates that the last page has been retrieved.
+    /// - If there are no leases returned it indicates that the previous page
+    ///   was the last page.
+    ///
+    /// @param lower_bound_address IPv6 address used as lower bound for the
+    /// returned range.
+    /// @param page_size maximum size of the page returned.
+    ///
+    /// @return Lease collection (may be empty if no IPv6 lease found).
+    virtual Lease6Collection
+    getLeases6(const asiolink::IOAddress& lower_bound_address,
+               const LeasePageSize& page_size) const;
+
     /// @brief Returns a collection of expired DHCPv4 leases.
     ///
     /// This method returns at most @c max_leases expired leases. The leases
@@ -316,7 +403,16 @@ public:
     ///
     /// @param lease4 The lease to be updated.
     ///
-    /// If no such lease is present, an exception will be thrown.
+    /// @throw NoSuchLease if there is no such lease to be updated.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the update
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and UPDATE with different expiration time.
     virtual void updateLease4(const Lease4Ptr& lease4);
 
     /// @brief Updates IPv6 lease.
@@ -326,16 +422,49 @@ public:
     ///
     /// @param lease6 The lease to be updated.
     ///
-    /// If no such lease is present, an exception will be thrown.
+    /// @throw NoSuchLease if there is no such lease to be updated.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the update
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and UPDATE with different expiration time.
     virtual void updateLease6(const Lease6Ptr& lease6);
 
-    /// @brief Deletes a lease.
+    /// @brief Deletes an IPv4 lease.
     ///
-    /// @param addr Address of the lease to be deleted. (This can be IPv4 or
-    ///        IPv6.)
+    /// @param lease IPv4 lease being deleted.
     ///
-    /// @return true if deletion was successful, false if no such lease exists
-    virtual bool deleteLease(const isc::asiolink::IOAddress& addr);
+    /// @return true if deletion was successful, false if no such lease exists.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the deletion
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and DELETE with different expiration time.
+    virtual bool deleteLease(const Lease4Ptr& lease);
+
+    /// @brief Deletes an IPv6 lease.
+    ///
+    /// @param lease IPv6 lease being deleted.
+    ///
+    /// @return true if deletion was successful, false if no such lease exists.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the deletion
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and DELETE with different expiration time.
+    virtual bool deleteLease(const Lease6Ptr& lease);
 
     /// @brief Deletes all expired-reclaimed DHCPv4 leases.
     ///
@@ -361,6 +490,7 @@ public:
     /// subnet.
     ///
     /// @param subnet_id identifier of the subnet
+    ///
     /// @return number of leases removed.
     virtual size_t wipeLeases4(const SubnetID& subnet_id);
 
@@ -370,11 +500,278 @@ public:
     /// subnet.
     ///
     /// @param subnet_id identifier of the subnet
+    ///
     /// @return number of leases removed.
     virtual size_t wipeLeases6(const SubnetID& subnet_id);
 
 private:
 
+    /// @name Internal methods called while holding the mutex in multi threading
+    /// mode.
+    ///@{
+
+    /// @brief Adds an IPv4 lease,
+    ///
+    /// @param lease lease to be added
+    ///
+    /// @result true if the lease was added, false if not
+    bool addLeaseInternal(const Lease4Ptr& lease);
+
+    /// @brief Adds an IPv6 lease.
+    ///
+    /// @param lease lease to be added
+    ///
+    /// @result true if the lease was added, false if not
+    bool addLeaseInternal(const Lease6Ptr& lease);
+
+    /// @brief Returns existing IPv4 lease for specified IPv4 address.
+    ///
+    /// @param addr An address of the searched lease.
+    ///
+    /// @return a pointer to the lease (or NULL if a lease is not found)
+    Lease4Ptr getLease4Internal(const isc::asiolink::IOAddress& addr) const;
+
+    /// @brief Gets existing IPv4 leases for specified hardware address.
+    ///
+    /// @param hwaddr hardware address of the client
+    /// @param collection lease collection
+    void getLease4Internal(const isc::dhcp::HWAddr& hwaddr,
+                           Lease4Collection& collection) const;
+
+    /// @brief Returns existing IPv4 lease for specified hardware address
+    ///        and a subnet
+    ///
+    /// @param hwaddr hardware address of the client
+    /// @param subnet_id identifier of the subnet that lease must belong to
+    ///
+    /// @return a pointer to the lease (or NULL if a lease is not found)
+    Lease4Ptr getLease4Internal(const HWAddr& hwaddr,
+                                SubnetID subnet_id) const;
+
+    /// @brief Gets existing IPv4 lease for specified client-id
+    ///
+    /// @param client_id client identifier
+    /// @param collection lease collection
+    void getLease4Internal(const ClientId& client_id,
+                           Lease4Collection& collection) const;
+
+    /// @brief Returns IPv4 lease for specified client-id/hwaddr/subnet-id tuple
+    ///
+    /// @param clientid client identifier
+    /// @param hwaddr hardware address of the client
+    /// @param subnet_id identifier of the subnet that lease must belong to
+    ///
+    /// @return a pointer to the lease (or NULL if a lease is not found)
+    Lease4Ptr getLease4Internal(const ClientId& clientid,
+                                const HWAddr& hwaddr,
+                                SubnetID subnet_id) const;
+
+    /// @brief Returns existing IPv4 lease for specified client-id
+    ///
+    /// @param clientid client identifier
+    /// @param subnet_id identifier of the subnet that lease must belong to
+    ///
+    /// @return a pointer to the lease (or NULL if a lease is not found)
+    Lease4Ptr getLease4Internal(const ClientId& clientid,
+                                SubnetID subnet_id) const;
+
+    /// @brief Gets all IPv4 leases for the particular subnet identifier.
+    ///
+    /// @param subnet_id subnet identifier.
+    /// @param collection lease collection
+    void getLeases4Internal(SubnetID subnet_id,
+                            Lease4Collection& collection) const;
+
+    /// @brief Returns all IPv4 leases for the particular hostname.
+    ///
+    /// @param hostname hostname in lower case.
+    /// @param collection lease collection
+    void getLeases4Internal(const std::string& hostname,
+                            Lease4Collection& collection) const;
+
+    /// @brief Gets all IPv4 leases.
+    ///
+    /// @param collection lease collection
+    void getLeases4Internal(Lease4Collection& collection) const;
+
+    /// @brief Returns range of IPv4 leases using paging.
+    ///
+    /// @param lower_bound_address IPv4 address used as lower bound for the
+    /// returned range.
+    /// @param page_size maximum size of the page returned.
+    /// @param collection lease collection
+    void getLeases4Internal(const asiolink::IOAddress& lower_bound_address,
+                            const LeasePageSize& page_size,
+                            Lease4Collection& collection) const;
+
+    /// @brief Returns existing IPv6 lease for a given IPv6 address.
+    ///
+    /// @param type specifies lease type: (NA, TA or PD)
+    /// @param addr An address of the searched lease.
+    ///
+    /// @return a pointer to the lease (or NULL if a lease is not found)
+    Lease6Ptr getLease6Internal(Lease::Type type,
+                                const isc::asiolink::IOAddress& addr) const;
+
+    /// @brief Returns existing IPv6 lease for a given DUID + IA + lease type
+    /// combination
+    ///
+    /// @param type specifies lease type: (NA, TA or PD)
+    /// @param duid client DUID
+    /// @param iaid IA identifier
+    /// @param collection lease collection
+    void getLeases6Internal(Lease::Type type,
+                            const DUID& duid,
+                            uint32_t iaid,
+                            Lease6Collection& collection) const;
+
+    /// @brief Returns existing IPv6 lease for a given DUID + IA + subnet-id +
+    /// lease type combination.
+    ///
+    /// @param type specifies lease type: (NA, TA or PD)
+    /// @param duid client DUID
+    /// @param iaid IA identifier
+    /// @param subnet_id identifier of the subnet the lease must belong to
+    /// @param collection lease collection
+    void getLeases6Internal(Lease::Type type,
+                            const DUID& duid,
+                            uint32_t iaid,
+                            SubnetID subnet_id,
+                            Lease6Collection& collection) const;
+
+    /// @brief Returns all IPv6 leases for the particular subnet identifier.
+    ///
+    /// @param subnet_id subnet identifier.
+    /// @param collection lease collection
+    void getLeases6Internal(SubnetID subnet_id,
+                            Lease6Collection& collection) const;
+
+    /// @brief Returns all IPv6 leases for the particular hostname.
+    ///
+    /// @param hostname hostname in lower case.
+    /// @param collection lease collection
+    void getLeases6Internal(const std::string& hostname,
+                            Lease6Collection& collection) const;
+
+    /// @brief Returns all IPv6 leases.
+    ///
+    /// @param collection lease collection
+    void getLeases6Internal(Lease6Collection& collection) const;
+
+    /// @brief Returns IPv6 leases for the DUID.
+    ///
+    /// @param duid client DUID
+    /// @param collection lease collection
+    void getLeases6Internal(const DUID& duid,
+                            Lease6Collection& collection) const;
+
+    /// @brief Returns range of IPv6 leases using paging.
+    ///
+    /// @param lower_bound_address IPv6 address used as lower bound for the
+    /// returned range.
+    /// @param page_size maximum size of the page returned.
+    /// @param collection lease collection
+    void getLeases6Internal(const asiolink::IOAddress& lower_bound_address,
+                            const LeasePageSize& page_size,
+                            Lease6Collection& collection) const;
+
+    /// @brief Returns a collection of expired DHCPv4 leases.
+    ///
+    /// @param [out] expired_leases A container to which expired leases returned
+    /// by the database backend are added.
+    /// @param max_leases A maximum number of leases to be returned. If this
+    /// value is set to 0, all expired (but not reclaimed) leases are returned.
+    void getExpiredLeases4Internal(Lease4Collection& expired_leases,
+                                   const size_t max_leases) const;
+
+    /// @brief Returns a collection of expired DHCPv6 leases.
+    ///
+    /// @param [out] expired_leases A container to which expired leases returned
+    /// by the database backend are added.
+    /// @param max_leases A maximum number of leases to be returned. If this
+    /// value is set to 0, all expired (but not reclaimed) leases are returned.
+    void getExpiredLeases6Internal(Lease6Collection& expired_leases,
+                                   const size_t max_leases) const;
+
+    /// @brief Updates IPv4 lease.
+    ///
+    /// @param lease4 The lease to be updated.
+    ///
+    /// @throw NoSuchLease if there is no such lease to be updated.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the update
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and UPDATE with different expiration time.
+    void updateLease4Internal(const Lease4Ptr& lease4);
+
+    /// @brief Updates IPv6 lease.
+    ///
+    /// @param lease6 The lease to be updated.
+    ///
+    /// @throw NoSuchLease if there is no such lease to be updated.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the update
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and UPDATE with different expiration time.
+    void updateLease6Internal(const Lease6Ptr& lease6);
+
+    /// @brief Deletes an IPv4 lease.
+    ///
+    /// @param lease IPv4 lease being deleted.
+    ///
+    /// @return true if deletion was successful, false if no such lease exists.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the deletion
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and DELETE with different expiration time.
+    bool deleteLeaseInternal(const Lease4Ptr& addr);
+
+    /// @brief Deletes an IPv6 lease.
+    ///
+    /// @param lease IPv6 lease being deleted.
+    ///
+    /// @return true if deletion was successful, false if no such lease exists.
+    ///
+    /// @note The current_cltt_ and current_valid_lft_ are used to maximize the
+    /// chance that only one thread or process performs an update or delete
+    /// operation on the lease by matching these values with the expiration time
+    /// data in the database.
+    /// @note For test purposes only, when persistence is disabled, the deletion
+    /// of the lease is performed only if the value matches the one received on
+    /// the SELECT query, effectively enforcing no update on the lease between
+    /// SELECT and DELETE with different expiration time.
+    bool deleteLeaseInternal(const Lease6Ptr& addr);
+
+    /// @brief Removes specified IPv4 leases.
+    ///
+    /// @param subnet_id identifier of the subnet
+    ///
+    /// @return The number of deleted leases
+    size_t wipeLeases4Internal(const SubnetID& subnet_id);
+
+    /// @brief Removed specified IPv6 leases.
+    ///
+    /// @param subnet_id identifier of the subnet
+    ///
+    /// @return The number of deleted leases
+    size_t wipeLeases6Internal(const SubnetID& subnet_id);
+    ///@}
+
     /// @brief Deletes all expired-reclaimed leases.
     ///
     /// This private method is called by both of the public methods:
@@ -484,7 +881,9 @@ public:
     /// - LFC Finish File: ".completed"
     /// - LFC PID File: ".pid"
     ///
-    /// See http://kea.isc.org/wiki/LFCDesign for details.
+    /// See
+    /// https://gitlab.isc.org/isc-projects/kea/wikis/designs/Lease-File-Cleanup-design
+    /// for details.
     ///
     /// @param file_name A base file name to which suffix is appended.
     /// @param file_type An LFC file type.
@@ -570,7 +969,9 @@ private:
     ///
     /// The <filename>.2, <filename>.1 and <filename>.completed are the
     /// products of the lease file cleanups (LFC).
-    /// See: http://kea.isc.org/wiki/LFCDesign for details.
+    /// See:
+    /// https://gitlab.isc.org/isc-projects/kea/wikis/designs/Lease-File-Cleanup-design
+    /// for details.
     ///
     /// @note: When the server starts up or is reconfigured it will try to
     /// read leases from the lease files using this method. It is possible
@@ -750,7 +1151,7 @@ private:
     /// the lease files.
     ///
     /// For the design and the terminology used in this description refer to
-    /// the http://kea.isc.org/wiki/LFCDesign.
+    /// the https://gitlab.isc.org/isc-projects/kea/wikis/designs/Lease-File-Cleanup-design.
     ///
     /// If the method finds that the %Lease File Copy exists it simply runs
     /// the @c kea-lfc application.
@@ -774,12 +1175,15 @@ private:
     ///
     /// DatabaseConnection object is used only for storing, accessing and
     /// printing parameter map.
-    DatabaseConnection conn_;
+    db::DatabaseConnection conn_;
 
     //@}
+
+    /// @brief Manager mutex
+    boost::scoped_ptr<std::mutex> mutex_;
 };
 
-}; // end of isc::dhcp namespace
-}; // end of isc namespace
+}  // namespace dhcp
+}  // namespace isc
 
 #endif // MEMFILE_LEASE_MGR_H
